#!/usr/bin/python

import paramiko
import socket
import os
import os.path
import select
import syslog
import time
import gettext
import stat
import io
import tempfile
import ConfigParser

#############################################################################
# Config defaults.  Can be overidden in /etc/pam-sshauth.conf
#############################################################################

config = {
    'SSH':        'ssh -YMNnq -o ServerAliveInterval=120 -S {socket} -p {port} -l {user} {host}',
    'UIDCMD':     'id -u',
    'GIDCMD':     'id -g',
    'PASSWDCMD':  'getent passwd $(id -u)',
    'GROUPSCMD':  'for G in $(id -G); do getent group $G; done',
    'EXTRAUSERS': '/var/lib/extrausers',
    'TMPDIR':     '/var/tmp'
    }

#############################################################################
# Constants
#############################################################################

SOCKET = 'PAM_SSHAUTH_SOCKET'
HOST   = 'PAM_SSHAUTH_HOST'
PORT   = 'PAM_SSHAUTH_PORT'
UID    = 'PAM_SSHAUTH_UID'
GID    = 'PAM_SSHAUTH_GID'

#############################################################################
# Globals
#############################################################################

pamh           = None
gdebug         = False
try_first_pass = False
first_time     = True

#############################################################################
# Gettext Initialization
#############################################################################

gettext.bindtextdomain('pam-sshauth', '/usr/share/pam-sshauth/lang')
gettext.textdomain('pam-sshauth')
_ = gettext.gettext

#############################################################################
# debug
#
# Syslogs infomation if gdebug global set
#############################################################################

def debug(var):
    if gdebug:
        syslog.openlog('pam-sshauth', syslog.LOG_PID, syslog.LOG_AUTH)
        syslog.syslog(var)
        syslog.closelog()

#############################################################################
# pam helpers message, error, and prompt
#############################################################################

def message(msg):
    try:
        pamh.conversation(pamh.Message(pamh.PAM_TEXT_INFO, msg))
    except pamh.exception as err:
        debug('Error: conversation returned: %s' % (err))
        return err.pam_result
    return pamh.PAM_SUCCESS

def error(msg):
    try:
        pamh.conversation(pamh.Message(pamh.PAM_ERROR_MSG, msg))
    except pamh.exception as err:
        debug('Error: conversation returned: %s' % (err))
        return err.pam_result
    return pamh.PAM_SUCCESS

def prompt(echo, msg):
    try:
        response = pamh.conversation(pamh.Message(echo, msg))
    except pamh.exception as err:
        debug('Error: conversation returned: %s' % (err))
        return (err.pam_result, None)
    return (pamh.PAM_SUCCESS, response.resp)

#############################################################################
# start_master_socket
#
# This does the standard double-fork-exec to launch an ssh master socket.
# It will then watch for the ssh process to ask for the password, and supply
# it as needed.  Uses forkpty to defeat ssh's attempt to only read input from
# /dev/tty.
#############################################################################

def start_master_socket():
    global pamh
    global config
    try:
        pid1 = os.fork()
    except:
        return

    if not pid1:
        os.setsid()
        try:
            pid, fd = os.forkpty()
        except:
            os._exit(1)

        if not pid:
            # In child, drop privs and spawn ssh tunnel
            ssh_tunnel = config['SSH'].format(socket = pamh.env[SOCKET],
                    user = pamh.get_user(), port = pamh.env[PORT],
                    host = pamh.env[HOST]).split(' ')

            os.setgid(int(pamh.env[GID]))
            os.setuid(int(pamh.env[UID]))

            if pamh.xdisplay:
                os.environ['DISPLAY'] = pamh.xdisplay

            os.execvp(ssh_tunnel[0], ssh_tunnel)
            os._exit(1)

        while True:
            rfds, wfds, xfds = select.select([fd], [], [])
            if fd in rfds:
                try:
                    # We received input, it's asking for a password, supply it
                    output = os.read(fd, io.DEFAULT_BUFFER_SIZE)
                    os.write(fd, pamh.authtok)
                    if not pamh.authtok.endswith('\n'):
                        os.write(fd, '\n')
                    break
                except OSError:
                    pass

        os.waitpid(pid, 0)
        os._exit(0)

#############################################################################
# ihandler
#
# Interactive handler function for the keyboard-interactive authentication
# method.  Needed for if you wish to correctly handle things like password
# expiry.
#############################################################################

def ihandler(title, instructions, prompt_list):
    global pamh
    global first_time
    global try_first_pass

    results = []
    if instructions:
        message(instructions)

    if first_time and try_first_pass:
        results.append(pamh.authtok)
        first_time = False
        return results

    for prompt in prompt_list:
        if prompt[1]:
            echo = pamh.PAM_PROMPT_ECHO_ON
        else:
            echo = pamh.PAM_PROMPT_ECHO_OFF
        pam_result, response = prompt(echo, prompt[0])
        if pam_result == pamh.PAM_SUCCESS:
            results.append(response)
            if echo == pamh.PAM_PROMPT_ECHO_OFF:
                pamh.authtok = response
    return results

#############################################################################
# parse_args
#
# parse the pam command line arguments.
#############################################################################

def parse_args(argv):
    global pamh
    global gdebug
    global try_first_pass
    pamh.env[HOST] = 'server'
    pamh.env[PORT] = '22'
    tunnel = False
    method = 'password'

    for arg in argv[1:]:
        if 'host=' in arg:
            pamh.env[HOST] = arg[5:]
        if 'port=' in arg:
            pamh.env[PORT] = arg[5:]
        if 'method=' in arg:
            method = arg[7:]
        if arg == 'tunnel':
            tunnel = True
        if arg == 'debug':
            gdebug = True
        if arg == 'try_first_pass':
            try_first_pass = True
    return method, tunnel

#############################################################################
# run_remote
#
# Run a remote command via the ssh transport.
#############################################################################

def run_remote(transport, command):
    debug('Running remote command: ' + command)
    text = ''
    channel = transport.open_channel(kind = 'session')
    channel.exec_command(command)
    while not channel.exit_status_ready():
        text += channel.recv(io.DEFAULT_BUFFER_SIZE)
    return text

#############################################################################
# merge
#
# Helper function to manage the nss-extrausers files.
#############################################################################

def merge(mergedict, db, mergelist):
    for line in mergelist:
        elem = line.split(':')
        if len(elem) > 3:
            if not line.endswith('\n'):
                line += '\n'
            mergedict[elem[2]] = line

#############################################################################
# update_nss
#
# Update the nss-extrausers database with passwd and group entries
#############################################################################

def update_nss(db, data):
    global config
    merged = {}
    fname = os.path.join(config['EXTRAUSERS'], db)
    debug('Updating %s' % (fname))
    try:
        f = open (fname, 'r')
        merge(merged, db, f)
        f.close()
    except:
        pass

    newlines = data.split('\n')
    merge(merged, db, newlines)

    f = open (fname, 'w+')
    f.writelines(merged.values())
    f.close()

    # make sure files have correct permissions
    os.chmod(fname, stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH)

#############################################################################
# get_userinfo
#
# Pull user geometry over the ssh transport, and add to nss-extrausers
#############################################################################

def get_userinfo(transport):
    global pamh
    global config
    # Get info
    pamh.env[UID] = run_remote(transport, config['UIDCMD'])
    pamh.env[GID] = run_remote(transport, config['GIDCMD'])
    passwdentry   = run_remote(transport, config['PASSWDCMD'])
    groupentry    = run_remote(transport, config['GROUPSCMD'])

    # We've gotten the passwd entry for our user from the remote system.
    # set up passwd and group entries for nss-extrausers

    update_nss('passwd', passwdentry)
    update_nss('group',  groupentry)

#############################################################################
# ssh_close
#
# Small helper function to close the ssh transport.
#############################################################################

def ssh_close(transport):
    # Paramiko automatically performs a socket close, so once we close
    # the transport, the socket is closed as well.

    if transport.is_active():
        transport.close()

#############################################################################
# auth_ki
#
# Keyboard-interactive authenticator
#############################################################################

def auth_ki(transport, username):
    global pamh

    try:
        transport.auth_interactive (username, ihandler)
    except paramiko.AuthenticationException as err:
        debug('Error: auth_interactive returned: %s' % (err))
    if not transport.is_authenticated():
        return pamh.PAM_AUTH_ERR
    return pamh.PAM_SUCCESS

#############################################################################
# auth_pw
#
# Password authenticator
#############################################################################

def auth_pw(transport, username):
    global pamh
    global try_first_pass
    global first_time

    password = ''

    if first_time and try_first_pass:
        password = pamh.authtok
        first_time = False
    else:
        pam_result, response = prompt(pamh.PAM_PROMPT_ECHO_OFF, _('Password: '))
        if pam_result != pamh.PAM_SUCCESS:
            ssh_close(transport)
            return pam_result
        password = response

    try:
        transport.auth_password(username, password)
    except paramiko.AuthenticationException as err:
        debug('Error: auth_password returned: %s' % (err))

    if not transport.is_authenticated():
        pam_result = error(_('Password incorrect.'))
        if pam_result != pamh.PAM_SUCCESS:
            return pam_result
        return pamh.PAM_AUTH_ERR

    pamh.authtok = password
    return pamh.PAM_SUCCESS

#############################################################################
# auth_pk
#
# Publickey authenticator
#############################################################################

def auth_pk(transport, username):
    global pamh

    keys = paramiko.Agent().get_keys()

    debug('Number of keys returned from agent: %d' % len(keys))
    for key in keys:
        try:
            transport.auth_publickey(username, key)
        except paramiko.AuthenticationException as err:
            debug('Error: auth_publickey returned: %s' % (err))
        if transport.is_authenticated():
            return pamh.PAM_SUCCESS

    return pamh.PAM_AUTH_ERR

#############################################################################
# ssh_auth
#
# Authenticate a pam user via an ssh connection to a remote machine.
#############################################################################

def ssh_auth(method, tunnel):
    global pamh
    global try_first_pass
    global first_time
    global config

    debug('Beginning SSH authentication')

    try:
        username = pamh.get_user()
    except pamh.exception as err:
        debug('Error: get_user returned: %s' % (err))
        return err.pam_result

    try:
        sock = socket.socket()
        sock.connect((pamh.env[HOST], int(pamh.env[PORT])))
    except socket.error as err:
        debug('Error: socket.connect returned: %s' % (err))
        return pamh.PAM_SYSTEM_ERR

    transport = paramiko.Transport(sock)

    transport.start_client()

    allowed_types = []
    try:
        transport.auth_none(username)
    except paramiko.BadAuthenticationType as err:
        allowed_types = err.allowed_types

    debug('Allowed auth types: %s' % (' '.join(allowed_types)))
    debug('Method = ' + method)

    for i in range(0, 3):
        pam_result = pamh.PAM_AUTH_ERR
        if ('keyboard-interactive' in allowed_types
            and method == 'keyboard-interactive'):
            pam_result = auth_ki(transport, username)
        if 'password' in allowed_types and method == 'password':
            pam_result = auth_pw(transport, username)
        if 'publickey' in allowed_types and method == 'publickey':
            pam_result = auth_pk(transport, username)
        if pam_result == pamh.PAM_SUCCESS:
            break

    if not transport.is_authenticated():
        debug('Authentication failed')
        ssh_close(transport)
        pam_result = error(_('Authentication failed.'))
        if pam_result != pamh.PAM_SUCCESS:
            return pam_result
        return pamh.PAM_AUTH_ERR

    debug('Authentication succeeded')

    if tunnel and os.geteuid() == 0:
        get_userinfo(transport)
        ssh_close(transport)
        tmpdir = tempfile.mkdtemp(dir = config['TMPDIR'])
        os.chown(tmpdir, int(pamh.env[UID]), int(pamh.env[GID]))
        pamh.env[SOCKET] = os.path.join(tmpdir, username)

        if not os.path.exists(pamh.env[SOCKET]):
            start_master_socket()

            # if this is an tunnel login, we NEED that master socket.  We'll
            # wait up to 10 seconds for it to appear.  If it doesn't,
            # return PAM_SYSTEM_ERR
            debug('Waiting for control socket %s...' % (pamh.env[SOCKET]))
            tick = 100
            while not os.path.exists(pamh.env[SOCKET]) and tick:
                time.sleep(.1)
                tick -= 1

            if not os.path.exists(pamh.env[SOCKET]):
                debug('Error: Control socket not created')
                return pamh.PAM_SYSTEM_ERR
            debug('Control socket created')

    ssh_close(transport)

    return pamh.PAM_SUCCESS

#############################################################################
# get_config
#
# Read the configfile (if it exists) and override commands
#############################################################################

def get_config():
    global config
    parser = ConfigParser.ConfigParser()
    parser.optionxform = str
    parser.read('/etc/pam-sshauth.conf')
    for section in parser.sections():
        for option in parser.options(section):
            text = parser.get(section, option)
            config[option] = text
            debug('config[%s] = %s' % (option, text))

#############################################################################
#
# pam_python entry points we handle.
#
#############################################################################

def pam_sm_authenticate(pamhandle, flags, argv):
    global pamh
    pamh = pamhandle
    method, tunnel = parse_args(argv)
    get_config()
    return ssh_auth(method, tunnel)

def pam_sm_setcred(pamhandle, flags, argv):
    return pamhandle.PAM_SUCCESS

def pam_sm_acct_mgmt(pamhandle, flags, argv):
    return pamhandle.PAM_SUCCESS

def pam_sm_open_session(pamhandle, flags, argv):
    return pamhandle.PAM_SUCCESS

def pam_sm_close_session(pamhandle, flags, argv):
    return pamhandle.PAM_SUCCESS
